@@ -1,5 +1,6 @@
 /*
- * Copyright 2015, 2016 Andrew Ayer
+ * Copyright 2015, 2016 Andrew Ayer <agwa@andrewayer.name>
+ * Copyright 2016, 2017 Chris Lamb <lamby@debian.org>
  *
  * This file is part of disorderfs.
  *
@@ -36,13 +37,13 @@ extern "C" {
 #include <vector>
 #include <random>
 #include <algorithm>
-#include <attr/xattr.h>
+#include <sys/xattr.h>
 #include <sys/types.h>
 #include <sys/syscall.h>
 #include <sys/file.h>
 #include <stddef.h>
 
-#define DISORDERFS_VERSION "0.5.2"
+#define DISORDERFS_VERSION "0.5.6"
 
 namespace {
 	std::vector<std::string>	bare_arguments;
@@ -67,7 +68,7 @@ namespace {
 	}
 
 	int wrap (int retval) { return retval == -1 ? -errno : 0; }
-	using Dirents = std::vector<std::string>;
+	using Dirents = std::vector<std::pair<std::string, ino_t>>;
 
 	// The libc versions of seteuid, etc. set the credentials for all threads.
 	// We need to set credentials for a single thread only, so call the syscalls directly.
@@ -269,7 +270,7 @@ int	main (int argc, char** argv)
 
 	// Add some of our own hard-coded FUSE options:
 	fuse_opt_add_arg(&fargs, "-o");
-	fuse_opt_add_arg(&fargs, "atomic_o_trunc,default_permissions"); // XXX: other mount options?
+	fuse_opt_add_arg(&fargs, "atomic_o_trunc,default_permissions,use_ino"); // XXX: other mount options?
 	if (config.multi_user) {
 		fuse_opt_add_arg(&fargs, "-o");
 		fuse_opt_add_arg(&fargs, "allow_other");
@@ -292,6 +293,12 @@ int	main (int argc, char** argv)
 	 * Initialize disorderfs_fuse_operations
 	 */
 
+	/*
+	 * Indicate that we should accept UTIME_OMIT (and UTIME_NOW) in the
+	 * utimens operations for "touch -m" and "touch -a"
+	 */
+	disorderfs_fuse_operations.flag_utime_omit_ok = 1;
+
 	disorderfs_fuse_operations.getattr = [] (const char* path, struct stat* st) -> int {
 		Guard g;
 		if (lstat((root + path).c_str(), st) == -1) {
@@ -424,11 +431,13 @@ int	main (int argc, char** argv)
 		if (!d) {
 			return -errno;
 		}
-		struct dirent	dirent_storage;
 		struct dirent*	dirent_p;
-		int		res;
-		while ((res = readdir_r(d, &dirent_storage, &dirent_p)) == 0 && dirent_p) {
-			dirents->emplace_back(dirent_p->d_name);
+		errno = 0;
+		while ((dirent_p = readdir(d)) != NULL) {
+			dirents->emplace_back(std::make_pair(dirent_p->d_name, dirent_p->d_ino));
+		}
+		if (errno != 0) {
+			return -errno;
 		}
 		if (config.sort_dirents) {
 			std::sort(dirents->begin(), dirents->end());
@@ -437,8 +446,8 @@ int	main (int argc, char** argv)
 			std::reverse(dirents->begin(), dirents->end());
 		}
 		closedir(d);
-		if (res != 0) {
-			return -res;
+		if (errno != 0) {
+			return -errno;
 		}
 
 		set_fuse_data<Dirents*>(info, dirents.release());
@@ -446,14 +455,17 @@ int	main (int argc, char** argv)
 	};
 	disorderfs_fuse_operations.readdir = [] (const char* path, void* buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info* info) {
 		Dirents&		dirents = *get_fuse_data<Dirents*>(info);
+		struct stat		st;
+		memset(&st, 0, sizeof(st));
 		if (config.shuffle_dirents) {
 			std::random_device	rd;
 			std::mt19937		g(rd());
 			std::shuffle(dirents.begin(), dirents.end(), g);
 		}
 
-		for (const std::string& dirent : dirents) {
-			if (filler(buf, dirent.c_str(), nullptr, 0) != 0) {
+		for (const auto dirent : dirents) {
+			st.st_ino = dirent.second;
+			if (filler(buf, dirent.first.c_str(), &st, 0) != 0) {
 				return -ENOMEM;
 			}
 		}
